////////////////////////////////////////////////////////
// Project  : FHS
//
// File     : ahist.c
// Purpose  : 2 step Adaptive Kernel Density Estimation
// Author   : Damir Krstinic
// Copyright: (c) Damir Krstinic
// Version  : v1.0
////////////////////////////////////////////////////////
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <limits.h>
#include <float.h>

#include "fhs.h"
#include <assert.h>

/*
 *
 * Adaptive PDF (colorspace)
 *
 */
float *createAHist3D(IplImage *img, IplImage **hlabels, float h0, int *nbins) {
  int i, j;
  float *idata = (float *)(img->imageData);
  float *dp;
  
  float hr = h0/RHO, hi, hi2; // Histogram bin size
  int hbins;                  // Num. of hist bins
  int ptbin[3], ptkey;        // pt bin in 3D and key

  int ri;
  int cks[3], cke[3], ckc[3];
  float d[3], di;
  double glev, gmean=0;
  int npop=0;
  
  // Minimal bounding rectangle
  float vmax[3] = {idata[0], idata[1], idata[2]};
  float vmin[3] = {idata[0], idata[1], idata[2]};

  for (i=0; i<img->height; i++) {
    idata = (float *)(img->imageData + i*img->widthStep);
    for (j=0; j<img->width; j++) {
      dp = (idata + j*img->nChannels);
      if (dp[0] > vmax[0])
	vmax[0] = dp[0];
      else if (dp[0] < vmin[0])
	vmin[0] = dp[0];

      if (dp[1] > vmax[1]) 
	vmax[1] = dp[1];
      else if (dp[1] < vmin[1])
	vmin[1] = dp[1];

      if (dp[2] > vmax[2]) 
	vmax[2] = dp[2];
      else if (dp[2] < vmin[2])
	vmin[2] = dp[2];
    }
  }

  for (i=0; i<3; i++) {
    nbins[i] = (int)ceil(1.001*(vmax[i]-vmin[i])/hr);
#ifdef DEBUG_OUT
    fprintf(stderr, "%1d: [%f, %f]; nbins = %d, h0 = %f, hr = %f\n",	\
	    i, vmin[i], vmax[i], nbins[i], h0, hr);
#endif
  }
  hbins = nbins[0] * nbins[1] * nbins[2];
#ifdef DEBUG_OUT
  fprintf(stderr, "hbins = %d\n", hbins);
#endif

  // Assertations
  assert(nbins[0] > 0);
  assert(nbins[1] > 0);
  assert(nbins[2] > 0);
  assert(hbins > 0);
  
  int *hist = calloc(hbins, sizeof(int));
  if (!hist) {
    perror("calloc");
    return NULL;
  }

  IplImage *labels = cvCreateImage(cvSize(img->width, img->height), \
				 IPL_DEPTH_32S, 1);
  if (!labels) {
    perror("cvCreateImage");
    return NULL;
  }
  int *ldata;

  for (i=0; i<img->height; i++) {
    idata = (float *)(img->imageData + i*img->widthStep);
    ldata = (int *)(labels->imageData + i*labels->widthStep);
    for (j=0; j<img->width; j++) {
      dp = (idata + j*img->nChannels);

      ptbin[0] = (int)floor((dp[0]-vmin[0])/hr);
      ptbin[1] = (int)floor((dp[1]-vmin[1])/hr);
      ptbin[2] = (int)floor((dp[2]-vmin[2])/hr);

      ptkey = ptbin[0] + ptbin[1]*nbins[0] +	\
	ptbin[2]*nbins[0]*nbins[1];
      
      // Assertations
      assert(ptbin[0] < nbins[0]); assert(ptbin[0] >= 0);
      assert(ptbin[1] < nbins[1]); assert(ptbin[1] >= 0);
      assert(ptbin[2] < nbins[2]); assert(ptbin[2] >= 0);
      assert(ptkey < hbins); assert(ptkey >= 0);
      
      ldata[j] = ptkey;
      // Hist
      hist[ptkey]++; // Centralna
    }
  }
  
  /* Geometric mean */
  for (i=0; i<hbins; i++) {
    if (hist[i] > 0) {
      gmean += log((double)hist[i]);
      npop++;
    }
  }
  gmean = exp(gmean/npop);
  glev = gmean;
  
#ifdef STD_OUT
    printf("gmean: %f\n", gmean); 
#endif
    
    float *fhist = calloc(hbins, sizeof(float));
    if (!fhist) {
      perror("calloc");
      return NULL;
    }
    
    npop = 0;
    for (ptbin[0]=0; ptbin[0]<nbins[0]; ptbin[0]++) {
      for (ptbin[1]=0; ptbin[1]<nbins[1]; ptbin[1]++) {
	for (ptbin[2]=0; ptbin[2]<nbins[2]; ptbin[2]++) {
	  ptkey = ptbin[0] + ptbin[1]*nbins[0] +	\
	    ptbin[2]*nbins[0]*nbins[1];
	  if (hist[ptkey] > 0) {
	    npop++;
	    hi = h0 * sqrt(glev/hist[ptkey]);
	    ri = (int)floor(hi/hr);
	    hi2 = hi * hi;
	    if (ri == 0) {
	      assert((double)fhist[ptkey] +				\
		     (double)hist[ptkey]/(hi2*hi) < FLT_MAX);
	      fhist[ptkey] += (float)hist[ptkey]/(hi2*hi);
	      continue;
	    }
	    cks[0] = MAX(ptbin[0]-ri, 0);
	    cke[0] = MIN(ptbin[0]+ri, nbins[0]-1);
	    cks[1] = MAX(ptbin[1]-ri, 0);
	    cke[1] = MIN(ptbin[1]+ri, nbins[1]-1);
	    cks[2] = MAX(ptbin[2]-ri, 0);
	    cke[2] = MIN(ptbin[2]+ri, nbins[2]-1);
	    for(ckc[2]=cks[2]; ckc[2]<=cke[2]; ckc[2]++) {
	      d[2] = (ckc[2] - ptbin[2])*hr;
	      d[2] = d[2] * d[2];
	      if (d[2] >= hi2)
		continue;
	      for(ckc[1]=cks[1]; ckc[1]<=cke[1]; ckc[1]++) {
		d[1] = (ckc[1] - ptbin[1])*hr;
		d[1] = d[1] * d[1];
		if (d[1] + d[2] >= hi2)
		  continue;
		for(ckc[0]=cks[0]; ckc[0]<=cke[0]; ckc[0]++) {
		  d[0] = (ckc[0] - ptbin[0])*hr;
		  d[0] = d[0] * d[0];
		  di = d[0] + d[1] + d[2];
#ifdef KERNEL_NORMAL
		  i = ckc[0] + ckc[1]*nbins[0] +			\
		    ckc[2]*nbins[0]*nbins[1];
		  assert((double)fhist[ptkey] +				\
			 (double)hist[ptkey]*exp(-di/(2*hi2))/(hi2*hi2*hi2) < FLT_MAX);
		  fhist[i] += (float)hist[ptkey]*exp(di/(2*hi2))/(hi2*hi2*hi2); 
#else
		  if (di < hi2) {
		    i = ckc[0] + ckc[1]*nbins[0] +			\
		      ckc[2]*nbins[0]*nbins[1];
		    assert((double)fhist[ptkey] +			\
			   (double)hist[ptkey]*(1-di/hi2)/(hi2*hi) < FLT_MAX);
		    fhist[i] += (float)hist[ptkey]*(1 - di/hi2)/(hi2*hi); 
		  }
#endif
		}
	      }
	    }
	  }
	}
      }
    }
    
#ifdef STD_OUT
    printf("Populated cells: %d\n", npop); fflush(stdout);
#endif
    free(hist);
    *hlabels = labels;
    return fhist; 
}
