////////////////////////////////////////////////////////
// Project  : FHS
//
// File     : cs_labeling.c
// Purpose  : Colorspace labeling
// Author   : Damir Krstinic
// Copyright: (c) Damir Krstinic
// Version  : v1.0
////////////////////////////////////////////////////////
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "fhs.h"
#include <assert.h>

static int cmp_modes(const void *mode1, const void *mode2) {
    cspaceClust *clust1 = (cspaceClust *)mode1;
    cspaceClust *clust2 = (cspaceClust *)mode2;

    if (clust1->cheight > clust2->cheight)
	return 1;
    else if (clust1->cheight < clust2->cheight)
	return -1;
    else
	return 0;
}

static int cid_sort(const void *c1, const void *c2) {
    return ((cspaceClust *)c1)->cid > ((cspaceClust *)c2)->cid ? 1 : -1;
}


/*
*
* BOTTOM-UP 3D recursive labeling
*
*/
float markCell3DBU(float *hist, float *hvalues, int k, int label, \
		 double eps1, int *nbins, cspaceClust *cinfo) {
    int ptbin[3];
    float histval;
    int nbstep;
    int nextcell = k;

    // Assertations
    assert(k < nbins[0]*nbins[1]*nbins[2]);
    assert(k >= 0);

    if (hist[k] <= 0) // Already has label!
	return hist[k];

    ptbin[0] = k%nbins[0];
    ptbin[1] = ((k-ptbin[0])/nbins[0])%nbins[1];
    ptbin[2] = (k-ptbin[0]-ptbin[1]*nbins[0])/(nbins[0]*nbins[1]);

    // Assertations
    assert(ptbin[0] < nbins[0]); assert(ptbin[0] >= 0);
    assert(ptbin[1] < nbins[1]); assert(ptbin[1] >= 0);
    assert(ptbin[2] < nbins[2]); assert(ptbin[2] >= 0);

    histval = hvalues[k];
    hist[k] = -label;

    // D0
    if (ptbin[0] > 0) {
	if (hvalues[k-1] >= histval && hist[k-1] != -label) {
	    nextcell = k-1;
	    histval = hvalues[k-1];
	}
    }
    if (ptbin[0] < nbins[0]-1) {
	if (hvalues[k+1] >= histval && hist[k+1] != -label) {
	    nextcell = k+1;
	    histval = hvalues[k+1];
	}
    }
    // D1
    if (ptbin[1] > 0) {
	if (hvalues[k-nbins[0]] >= histval && hist[k-nbins[0]] != -label) {
	    nextcell = k-nbins[0];
	    histval = hvalues[k-nbins[0]];
	}
    }
    if (ptbin[1] < nbins[1]-1) {
	if (hvalues[k+nbins[0]] >= histval && hist[k+nbins[0]] != -label) {
	    nextcell = k+nbins[0];
	    histval = hvalues[k+nbins[0]];
	}
    }
    // D2
    nbstep = nbins[0]*nbins[1];
    if (ptbin[2] > 0) {
	if (hvalues[k-nbstep] >= histval && hist[k-nbstep] != -label) {
	    nextcell = k-nbstep;
	    histval = hvalues[k-nbstep];
	}
    }
    if (ptbin[2] < nbins[2]-1) {
	if (hvalues[k+nbstep] >= histval && hist[k+nbstep] != -label) {
	    nextcell = k+nbstep;
	    histval = hvalues[k+nbstep];
	}
    }
    
    // Next step
    if (nextcell == k) { // Local Maximum
	if (histval >= eps1) { 
	    hist[k] = -label;
	    cinfo[label].cid     = label;
	    cinfo[label].label   = label;
	    cinfo[label].cheight = histval;
	    cinfo[label].cbin[0] = ptbin[0];
	    cinfo[label].cbin[1] = ptbin[1];
	    cinfo[label].cbin[2] = ptbin[2];
	} else {
	    hist[k] = 0;
	}
    } else {
	hist[k] = markCell3DBU(hist, hvalues, nextcell, label, eps1, nbins, \
	    cinfo);
    }

    return hist[k];
}


/*
*
*  BOTTOM-UP 3D labeling (colorspace labels)
*
*/
#define CSTART 200
#define CDIST  3
#define CFACT  1.6

int assignLabels3DBU(float *hist, int *nbins, int eps1, int eps0) {
  int k;
  int next_label = 1;
#ifdef STD_OUT
  int cremoved=0;
#endif
  int hbins = nbins[0]*nbins[1]*nbins[2];
  int cmem=CSTART;
  int ck1, clab, cdist;
  int ndist;
  double gmean = 0;
  int npop = 0;
  double lev0, lev1;

  float *hvalues = malloc(hbins*sizeof(float));
  cspaceClust *cinfo = calloc(CSTART, sizeof(cspaceClust));
  
  if (!hvalues || !cinfo) {
      perror("calloc");
      return -1;
  }
  memcpy(hvalues, hist, hbins*sizeof(float));

  /* Geometric mean */
  for (k=0; k<hbins; k++) {
      if (hist[k] > 0) {
	  gmean += log((double)hist[k]);
	  npop++;
      }
  }
  gmean = exp(gmean/npop);
  lev0 = eps0 * gmean; lev1 = eps1*gmean;

  k=0;
  while (k<hbins) {
      if (hist[k] > lev0) { 
	  hist[k] = markCell3DBU(hist, hvalues, k, next_label, lev1, nbins, \
	      cinfo);
	  if (hist[k] == -next_label) {
	      next_label++;
	  }
  
	  if (next_label == cmem) {
	      cmem *= 2;
	      cinfo = realloc(cinfo, cmem*sizeof(cspaceClust));
		    if (!cinfo) {
			perror("realloc");
			return -1;
		    }
	  }
#ifdef DEBUG_OUT
	  printf("[%9d] label assigned: %f\n", k, hist[k]);
#endif
      }

      while (k < hbins && hist[k] <= lev0)
	  k++;
  }

#ifdef STD_OUT
  printf("colorspace: %9d labels ... ", next_label);
#endif

  cdist = CDIST;
  qsort(&cinfo[1], next_label-1, sizeof(cspaceClust), cmp_modes);
  for (k=1; k<next_label; k++) {
      for (ck1=k+1; ck1<next_label; ck1++) {
	  ndist = abs(cinfo[k].cbin[0]-cinfo[ck1].cbin[0]) + \
	      abs(cinfo[k].cbin[1]-cinfo[ck1].cbin[1]) + \
	      abs(cinfo[k].cbin[2]-cinfo[ck1].cbin[2]);
	  if (ndist <= cdist) {
	      cinfo[k].label = cinfo[ck1].label;
	  }
      }
#ifdef STD_OUT
      if (cinfo[k].label != cinfo[k].cid)
	  cremoved++;
#endif
  }
  qsort(&cinfo[1], next_label-1, sizeof(cspaceClust), cid_sort);

#ifdef STD_OUT
  printf("%9d joined\n", cremoved);
#endif

  for (k=0; k<hbins; k++) {
      assert(hist[k] == (int)hist[k]);
      clab = cinfo[-(int)hist[k]].label;
      while(cinfo[clab].label != clab)
	  clab = cinfo[clab].label;

      hist[k] = clab;
  }

  free(hvalues);
  free(cinfo);

  return(next_label);
}
