////////////////////////////////////////////////////////
// Project  : FHS
//
// File     : img_labeling.c
// Purpose  : Image labeling
// Author   : Damir Krstinic
// Copyright: (c) Damir Krstinic
// Version  : v1.0
////////////////////////////////////////////////////////
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <sys/time.h>
#include <string.h>

#include "fhs.h"
#include <assert.h>


float *hist;
int npix;

IplImage *imglabels;
IplImage *hlabels;
IplImage *img;

#define START 0
#define LEFT  1
#define UP    2
#define RIGHT 3
#define DOWN  4
#define STOP  5

static inline float edge_weight(float *luva, float *luvb) {
    return sqrt((luva[0]-luvb[0]) * (luva[0]-luvb[0]) + \
		(luva[1]-luvb[1]) * (luva[1]-luvb[1]) + \
		(luva[2]-luvb[2]) * (luva[2]-luvb[2]));
}

static int cmp_edges(const void *edge1, const void *edge2) {
    einfo *e1 = (einfo *)edge1;
    einfo *e2 = (einfo *)edge2;
    if (e1->w > e2->w)
	return 1;
    else if (e1->w < e2->w)
	return -1;
    else
	return 0;
}

static int find_lab(clustStat *clstat, int l) {
    int nl = l;
    while (nl != clstat[nl].label)
	nl = clstat[nl].label;

    clstat[l].label = nl;
    return nl;
}

static void join_lab(clustStat *clstat, int la, int lb) {
    assert(la != lb);
    if (clstat[la].numPix > clstat[lb].numPix) {
	clstat[lb].label = la;
	clstat[la].numPix += clstat[lb].numPix;
    } else {
	clstat[la].label = lb;
	clstat[lb].numPix += clstat[la].numPix;
    }
}

/*
 *
 * Adding new edge
 *
 */
int checkEdge(egraph *gedg, int ia, int ja, char neighbour) {
  int ka = ia * img->width + ja;
  int ib = ia, jb = ja;

  switch (neighbour) {
  case LEFT:
    jb = ja - 1;
    break;
  case UP:
    ib = ia - 1;
    break;
  case RIGHT:
    jb = ja + 1;
    break;
  case DOWN:
    ib = ia + 1;
    break;
  }

  int *ladata = (int *)(imglabels->imageData + ia * imglabels->widthStep);
  int *lbdata = (int *)(imglabels->imageData + ib * imglabels->widthStep);

  int kb = ib * img->width + jb;

  int la = ladata[ja];
  int lb = lbdata[jb];
  int eel = gedg->size - 1;
  float w;
  int add_edge = 0;
  
  assert(la != lb);
  
  float *cola = (float *)(img->imageData + ia*img->widthStep);
  float *colb = (float *)(img->imageData + ib*img->widthStep);

  int *hladata = (int *)(hlabels->imageData + ia * hlabels->widthStep);
  int *hlbdata = (int *)(hlabels->imageData + ib * hlabels->widthStep);

  if (hist[hladata[ja]] == hist[hlbdata[jb]])
    w = 0;
  else
    w = edge_weight(&cola[ja*img->nChannels], &colb[jb*img->nChannels]);
  
  if (eel == -1) // Prvi rub (gedg->size == 0)!
    add_edge = 1;
  else {
    while(gedg->edges[eel].la == la && gedg->edges[eel].lb != lb && eel >= 0) 
      eel--;
    
    if (eel == -1) // Ne postoji
      add_edge = 1;
    else if (gedg->edges[eel].la != la) // Ne postoji
      add_edge = 1;
    else {  // Vec postoji, gledam tezinu
      assert(gedg->edges[eel].la == la && gedg->edges[eel].lb == lb);
      if (gedg->edges[eel].w > w)
	gedg->edges[eel].w = w;
    }
  }
  
  if (add_edge) {
    gedg->edges[gedg->size].la = la; // dodajem novi
    gedg->edges[gedg->size].lb = lb;
    gedg->edges[gedg->size].w  = w;
    gedg->size++;
    if (gedg->size == gedg->memsize) {
      gedg->memsize *= 2;
      gedg->edges = realloc(gedg->edges,			\
			    gedg->memsize*sizeof(einfo));
      if (!gedg->edges) {
	perror("realloc");
	exit(1);
      }
    }
  }
  
  return 0;
}

/*
 *
 * Mark pixel including spatial information
 *
 */

void markPix(int i, int j, int label, \
	     clustStat *statptr, egraph *gedg, char *path, int pstep) {
  int next_i = i, next_j = j;
  int next_step;

  float *idata = (float*)(img->imageData + i*img->widthStep);
  idata = idata + j * img->nChannels;
  // Assertations
  assert(pstep >= 0);
  assert(pstep < npix);
  assert(path[pstep] >= 0);
  assert(path[pstep] <= 5);
  
  assert(i>=0 && i < img->height);
  assert(j>=0 && j < img->width);
  
  int *ldata = (int *)(imglabels->imageData + i * imglabels->widthStep);
  int *lcheck;

  int *hldata = (int *)(hlabels->imageData + i * hlabels->widthStep);
  int *hlcheck;

  if (path[pstep] == START) {
    assert(ldata[j] == 0);
    assert(label > 0);
    ldata[j] = label;
    statptr->luv[0] += idata[0];
    statptr->luv[1] += idata[1];
    statptr->luv[2] += idata[2];
    statptr->numPix++;
    
    path[pstep] = LEFT;
  }

  if (path[pstep] == LEFT) {
    if (j > 0) {
      if (ldata[j-1] == 0) { // Neoznacena
	if (hist[hldata[j-1]] == hist[hldata[j]]) { // Isti
	  next_j = j-1;
	} else { // Nije isti clust. u spektralnoj domeni
	  path[pstep] = UP;
	}
      } else {     // Vec oznacena
	assert(ldata[j-1] > 0);
	if (ldata[j-1] != label) {
	  checkEdge(gedg, i, j, LEFT);
	}
	path[pstep] = UP;
      }
    } else        // Prva u redu, ne moze lijevo 
      path[pstep] = UP;
    
    assert((path[pstep] == UP && next_i == i && next_j == j) ||		\
	   (path[pstep] == LEFT && next_i == i && next_j == j-1));
  }
  
  if (path[pstep] == UP) {
    if (i > 0) {
      lcheck = (int *)(imglabels->imageData + \
		     (i-1) * imglabels->widthStep);
      if (lcheck[j] == 0) { // Neoznacena
	hlcheck = (int *)(hlabels->imageData + (i-1) * hlabels->widthStep);
	if (hist[hlcheck[j]] == hist[hldata[j]]) { // Isti
	  next_i = i-1;
	} else { // Nije isti clust. u spektralnoj domeni
	  path[pstep] = RIGHT;
	}
      } else {       // Vec oznacena
	assert(lcheck[j] > 0);
	if (lcheck[j] != label) {
	  checkEdge(gedg, i, j, UP);
	}
	path[pstep] = RIGHT;
      }
    } else   // Prvi red, ne moze gore
      path[pstep] = RIGHT;
    
    assert((path[pstep] == RIGHT && next_i == i && next_j == j) || \
	   (path[pstep] == UP && next_i == i-1 && next_j == j));
  }
  
  if (path[pstep] == RIGHT) {
    if (j < img->width - 1) {
      if (ldata[j+1] == 0) { // Neoznacena
	if (hist[hldata[j+1]] == hist[hldata[j]]) { // Isti
	  next_j = j+1;
	} else { // Nije isti clust. u spektralnoj domeni
	  path[pstep] = DOWN;
	}
      } else {
	assert(ldata[j+1] > 0);
	if (ldata[j+1] != label) {
	  checkEdge(gedg, i, j, RIGHT);
	}
	path[pstep] = DOWN;
      }
    } else 
      path[pstep] = DOWN;
    
    assert((path[pstep] == DOWN && next_i == i && next_j == j) || \
	   (path[pstep] == RIGHT && next_i == i && next_j == j+1));
  }
  
  if (path[pstep] == DOWN) {
    if (i < img->height - 1) {
      lcheck = (int *)(imglabels->imageData + \
			     (i+1) * imglabels->widthStep);
      if (lcheck[j] == 0) { // Neoznacena
	hlcheck = (int *)(hlabels->imageData + (i+1) * hlabels->widthStep);
	if (hist[hlcheck[j]] == hist[hldata[j]]) { // Isti
	  next_i = i+1;
	} else { // Nije isti clust. u spektralnoj domeni
	  path[pstep] = STOP;
	}
      } else {
	assert(lcheck[j] > 0);
	if (lcheck[j] != label) {
	  checkEdge(gedg, i, j, DOWN);
	}
	path[pstep] = STOP;
      }
    } else
      path[pstep] = STOP;
    
    assert((path[pstep] == STOP && next_i == i && next_j == j) || \
	   (path[pstep] == DOWN && next_i == i+1 && next_j == j));
  }
  
  if (path[pstep] == STOP) {
    assert(next_i == i && next_j == j);
    next_step = pstep-1; // Idem nazad
    if (next_step == -1) // Vratili se na prvog
      return;
    switch (path[next_step]-1) {
    case LEFT:
      next_i = i; next_j = j+1;
      break;
    case UP:
      next_i = i+1; next_j = j;
      break;
    case RIGHT:
      next_i = i; next_j = j-1;
      break;
    case DOWN:
      next_i = i-1; next_j = j;
      break;
    }
  } else {
    assert(path[pstep] > 0);
    assert(path[pstep] < 5);
    assert(next_i != i || next_j != j);
    path[pstep]++;
    next_step = pstep+1;
    path[next_step] = START;
  }
  
  markPix(next_i, next_j, label, statptr, gedg, path, next_step);
}
 
/*
 *
 * Grid labels to pixel labels including spatial information
 *
 */
#define C_NUM    4096

IplImage *gridLabels2img(IplImage *in_img, float *in_hist, \
			 IplImage *in_hlabels,				\
			 int *segments, int hbins, int minpts) {
    int i, j;
    int isegments;
    int next_label = 1, clsize = C_NUM;
    int la, lb;
    clustStat *statptr;
    clustStat *clstat; //  = malloc(C_NUM*sizeof(clustStat));
    einfo *edge;
    egraph gedg;
    int *ctable; // Tablica konverzija za konacne segmente na slici
    char *path;

#ifdef TIMES_OUT
    struct timeval time1, time2;
    int sectmp, msectmp;
#endif

    img = in_img;

    hist = in_hist;
    hlabels = in_hlabels;

    npix = img->height * img->width;

    imglabels = cvCreateImage(cvSize(img->width, img->height), \
				 IPL_DEPTH_32S, 1);
    if (!imglabels) {
	perror("cvCreateImage)");
	return NULL;
    }
    cvZero(imglabels);

    clstat = malloc(C_NUM*sizeof(clustStat));
    if (!clstat) {
	perror("malloc (clstat)");
	return NULL;
    }

#ifdef TIMES_OUT
    gettimeofday(&time1, NULL);
#endif

    path = malloc(npix*sizeof(char));
    if (!path) {
	perror("malloc (path)");
	return NULL;
    }

#ifdef DEBUG_OUT
    printf("Starting imgLabeling\n");
    fflush(stdout);fsync(STDOUT_FILENO);
#endif

    gedg.size    = 0;
    gedg.memsize = 4*C_NUM;
    gedg.edges   = malloc(gedg.memsize*sizeof(einfo));
    if (!gedg.edges) {
	perror("malloc");
	return NULL;
    }

    int *ldata;
    for (i=0; i<img->height; i++) {
      ldata = (int *)(imglabels->imageData + i * imglabels->widthStep);
      for (j=0; j<img->width; j++) {
	if (ldata[j] == 0) { // Nelabelirano
	  // Cluster statistics
	  statptr = &clstat[next_label];
	  statptr->label  = next_label;   // Cluster label
	  statptr->i = i;                 // Cluster representative
	  statptr->j = j;                 // coordinates
	  statptr->numPix = 0;            // Number of pixels
	  statptr->luv[0] = statptr->luv[1] = statptr->luv[2] = 0;
#ifdef DEBUG_OUT
	  printf("starting markPix([%3d, %3d]: l %9d) ... \n",	\
		 i, j, next_label); 
	  fflush(stdout);fsync(STDOUT_FILENO);
#endif
	  path[0] = START;
	  markPix(i, j, next_label, statptr, &gedg, path, 0); 
#ifdef DEBUG_OUT
	  printf("markPix OK\n"); 
	  fflush(stdout);fsync(STDOUT_FILENO);
#endif
	  statptr->luv[0] = statptr->luv[0]/statptr->numPix;
	  statptr->luv[1] = statptr->luv[1]/statptr->numPix;
	  statptr->luv[2] = statptr->luv[2]/statptr->numPix;
	  
	  /* Asertacije */
	  assert(ldata[j] == statptr->label);
	  assert(ldata[j] == clstat[ldata[j]].label);
	  /* ********** */
	  
	  next_label++;
	  if (next_label == clsize) {
	    clsize *= 2;
	    clstat = realloc(clstat, clsize*sizeof(clustStat));
	    if (!clstat) {
	      perror("realloc");
	      return NULL;
	    }
	  }
	}
      }
    }
    
#ifdef STD_OUT
    printf(" \t %d image segments\n", next_label);
    printf(" \t %d graph edges (memsize %d)\n", gedg.size, gedg.memsize);
    fflush(stdout);fsync(STDOUT_FILENO);
#endif

#ifdef TIMES_OUT
    gettimeofday(&time2, NULL);
    sectmp = time2.tv_sec - time1.tv_sec;
    msectmp = (time2.tv_usec - time1.tv_usec)/1000;
    if (msectmp < 0) {
	sectmp--;
	msectmp += 1000;
    }
    fprintf(stdout, "markPix: \t\t %3d:%03d\n", \
	    sectmp, msectmp); fflush(stdout);

    memcpy(&time1, &time2, sizeof(struct timeval));
#endif

    isegments = next_label;

    // Postprocesiranje malih segmenata
    qsort(gedg.edges, gedg.size, sizeof(einfo), cmp_edges);

#ifdef TIMES_OUT
    gettimeofday(&time2, NULL);
    sectmp = time2.tv_sec - time1.tv_sec;
    msectmp = (time2.tv_usec - time1.tv_usec)/1000;
    if (msectmp < 0) {
	sectmp--;
	msectmp += 1000;
    }
    fprintf(stdout, "sortiranje: \t\t %3d:%03d\n", \
	    sectmp, msectmp); fflush(stdout);

    memcpy(&time1, &time2, sizeof(struct timeval));
#endif

#ifdef STD_OUT
    printf("Graph based region growing \t OK\n");
    printf(" \t %d image segments\n", isegments);
    fflush(stdout);fsync(STDOUT_FILENO);
#endif

    for (i=0; i<gedg.size; i++) {
	edge = &gedg.edges[i];
	la = find_lab(clstat, edge->la);
	lb = find_lab(clstat, edge->lb);
	if (la != lb) {
	    if (clstat[la].numPix < minpts || clstat[lb].numPix < minpts) {
		join_lab(clstat, la, lb);
		isegments--;
	    }
	    
	}
    }

#ifdef STD_OUT
    printf("Small segments removed \t OK\n");
    printf(" \t %d image segments\n", isegments);
    fflush(stdout);fsync(STDOUT_FILENO);
#endif

    for (i=0; i<img->height; i++) {
      ldata = (int *)(imglabels->imageData + i * imglabels->widthStep);
      for (j=0; j<img->width; j++) {
	assert(ldata[j] > 0);
	assert(ldata[j] < next_label);
	ldata[j] = find_lab(clstat, ldata[j]);
      }
    }

    //////////////////
    if (!(ctable=calloc(isegments, sizeof(int)))) {
	perror("calloc");
	return NULL;
    }

    for (i=0; i<img->height; i++) {
      ldata = (int *)(imglabels->imageData + i * imglabels->widthStep);
      for (j=0; j<img->width; j++) {
	next_label = 0;
	while(ctable[next_label] != 0 && ctable[next_label] != ldata[j])
	    next_label++;
	
	assert(next_label < isegments);
	if (ctable[next_label] == 0) 
	  ctable[next_label] = ldata[j];

	ldata[j] = next_label;
      }
    }
    free(ctable);
    //////////////////
    isegments--;

#ifdef TIMES_OUT
    gettimeofday(&time2, NULL);
    sectmp = time2.tv_sec - time1.tv_sec;
    msectmp = (time2.tv_usec - time1.tv_usec)/1000;
    if (msectmp < 0) {
	sectmp--;
	msectmp += 1000;
    }
    fprintf(stdout, "spajanje malih: \t %3d:%03d\n", \
	    sectmp, msectmp); fflush(stdout);

    memcpy(&time1, &time2, sizeof(struct timeval));
#endif

    free(gedg.edges);
    free(clstat);
    
    *segments = isegments;

    return imglabels;
}
